#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <string>
#include <stl.h>
#include <multimap.h>
#ifdef HAVE_HASH_MAP
#include <hash_map>
#else
#include <multimap.h>
#endif
#if defined(HAVE_STRERROR) && defined(STDC_HEADERS)
#include <string.h>
#else
extern const char * const sys_errlist[];
extern const int sys_nerr;

inline const char *strerror(int e) {
    if ( e >=0 && e < sys_nerr ) return sys_errlist[e];
    else return "";
}
#endif

#ifdef HAVE_SYS_ERRNO_H
#include <errno.h>
#else
extern int errno;
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#else
extern "C" {
    int getopt(int, char * const *, const char *);
    int close(int);
}
#endif

#ifndef OPTARG_DEFINED
extern char *optarg;
extern int optind;
#endif

#ifdef HAVE_MMAP
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#define O_RDONLY 0x2
extern "C" {
    int open(const char *, int, ...);
}
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#endif

const int maxphonelen=10;
const char *opt = "q:z:f:L:l:n:vr";

#ifndef DEFAULT_DICT
#define DEFAULT_DICT "/usr/share/dict/words"
#endif
const char *default_dict = DEFAULT_DICT;

extern char *version;

#ifdef HAVE_HASH_MAP
// A functor for hashing strings - it is an adapter to get to the
// standard library char * hash function.
class stringhash : public unary_function<const string&, size_t> {
private:
    hash<const char *> h;
public:
    size_t operator()(const string& s) const {
	return h(s.c_str());
    }
};

typedef hash_multimap<string,string,stringhash> wordmap;
#else
// "less than" total order for strings - needed for maps
class stringless : public binary_function<const string&, const string&, bool> {
public:
    bool operator()(const string& s1, const string& s2) const {
	return (s1 < s2);
    }
};

typedef multimap<string,string,stringless> wordmap;
#endif
typedef list<string> stringlist;
typedef pair<wordmap::const_iterator, wordmap::const_iterator> wordmaprange;

#include "wordmatrix.h"

inline void fatal(const char *s) {
/* Terminate the program with an error message */
    cerr << s ;
    if ( errno ) cerr << ": " << strerror(errno);
    cerr << endl;

    exit(20);
}

void phone2words(const string& s, const string& prefix, unsigned int pos,
		wordmatrix &wp, stringlist& result) {
/* Recursive function to print the matches for this phone number.
 String is the phone number, prefix is the current prefix of the
 number, pos is the place to start our additions from.  wp is the
 wordmatrix - the function from position and length in s to lists of
 valid words with that length that appear at that point.  The
 recursion builds strings by setting the prefix to the longest valid
 word found at this position and recursing, then repeating for each
 smaller substring this level finds.  Once all the valid words are
 exhausted, it recurses by adding the digit at pos to the prefix and
 calling itself on the next position.  The recursion terminates when
 pos is at the end of s. */
    string pre;
    wordmap::const_iterator si;

    if ( pos == s.length() ) {
	result.push_back(prefix);
	return;
    }

    for ( int i = s.length() - pos; i > 0; i-- ) {
	for ( si = wp.get(pos, i).first; si != wp.get(pos, i).second; si++) {
	    pre = prefix;
	    if (pos) pre += " ";
	    pre += (*si).second;
	    phone2words(s,pre,pos+i,wp,result);
	}
    }
    pre = prefix;
    pre += s[pos];
    phone2words(s,pre,pos+1,wp,result);
}

inline bool tonumber(const string& s, char *num, const char*cm) {
// translate s to a number using the char map.  The caller must ensure
// that num points to a buffer of size at least s.length() +1.  If the
// mapping is complete true is returned; otherwise num contains an
// indeterminate string.
	for ( unsigned int i = 0; i < s.length(); i++ ) {
	    if ( isalpha(s[i])) 
		num[i] = cm[tolower(s[i])-'a'];
	    else	
		if (isdigit(s[i])) num[i] = s[i];
		else return 0;
	}
	num[s.length()] = '\0';
	return 1;
}

bool filemap(const char *file, wordmap& wm, const char *cm,
	       const unsigned int maxl, const unsigned int minl) {
/* Read file and insert an entry into wm for each word in the
   file. Words longer than maxl or shorter than minl are discarded.
   The function creates a key by using the lettermapping in cm to
   produce the string of digits that would make this word, and enters
   it into the wm multimap.  Words that contain non alphabetic
   characters are ignored.  The use of a char * for the key is a flat
   out hack for speed. To really speed it up, I should convert all the
   word operations to operations on char*'s, but I wrote this to play
   with standard strings, so...*/

    ifstream in(file);	// the input stream
    string word;	// The next word
    char *key;		// The key for that word

    if ( !in ) return 0;
    key = new char[maxl+1]; 

    while ( in >> word ) {
	if ( word.length() > maxl || word.length() < minl ) continue;
	if ( tonumber(word, key, cm) ) {
	    const wordmap::value_type p = wordmap::value_type(key,word);
	    wm.insert(p);
	}
    }
    delete [] key;
    return 1;
}


void init_cmap(char* cm) {
/* Initialize the array used to map between word and key letters.
   Because q and z are not on the dial the caller passes specific
   initializers in for those entries.  The mapping generated is for
   the standard US phone pad. When the funtion returns cm is an array
   that maps char-'a' -> [0-9] for char a lower case alpha
   character.*/
    char cc;	// The key number for this letter (other than q and z)
    int i;	// The number of times the current number has been used

    cc = '2';
    i = 0;
    for ( char c = 'a'; c <='z'; c++ )
	switch ( c ) {
	    case 'q' :
	    case 'z' :
		cm[c-'a'] = '-';
		break;
	    default:
		cm[c-'a']=cc;
		if ( ++i == 3 ) i=0, cc++;
		break;
	}
}

inline void cmap_map(char *cm, char from, char to) {
    cm[from-'a'] = to;
}

int main(int argc, char **argv) {
    wordmap wm;		// The map of numbers to words
    wordmaprange w;	// The list of words corresponding to this substring
    stringlist words;	// The words from this number
    char cm[26];	// The letter to number mapping
    string s;		// The current phone number to decode
    unsigned int start;	// index of starting point in this string
    int len;		// length of substring being considered
    int c;		// To parse opts
    unsigned int minl = 3; // The minimum word length in the mapping
    unsigned int maxl = maxphonelen;// The maximum word length in the mapping
    unsigned int pl=maxphonelen;// The maximum phone number that will be input
    bool makenumber =0;	// translate a phrase to a number
    int i;		// Scratch

    init_cmap(cm);
    while ( (c = getopt(argc,argv,opt)) != -1 ) 
	switch (c) {
	    case 'r':
		makenumber = 1;
		break;
	    case 'v':
		cerr << version << endl;
		exit(0);
	    case 'n':
		pl = atoi(optarg);
		if ( pl < 1) pl = maxphonelen;
		break;
	    case 'q':
		i = atoi(optarg);
		if ( i < 0 || i > 9 ) break;
		cmap_map(cm, 'q', i + '0');
		break;
	    case 'z':
		i = atoi(optarg);
		if ( i < 0 || i > 9 ) break;
		cmap_map(cm, 'z', i + '0');
		break;
	    case 'f':
		if ( !filemap(optarg,wm,cm, maxl, minl) )
		    cerr << "Can't open " << optarg << endl;
		break;
	    case 'l':
		minl = atoi(optarg);
		if ( minl < 1 ) minl = 1;
		break;
	    case 'L':
		maxl = atoi(optarg);
		if ( maxl > pl) maxl = pl;
		break;
	    case '?':
	    default:
		fatal("Bad argument");
		break;
	}

    if ( makenumber ) {
	for ( i = optind; i < argc ; i++ ) {
	    char *num = new char[pl+1];

	    if ( !num ) fatal("Can't allocate buffer");
	    s = argv[i];

	    if ( s.length() > pl ) s = s.substr(0,pl);

	    if ( !tonumber(s,num,cm) )
		cout << "No translation for " << s << endl;
	    else
		cout << num << endl;
	    if ( i+1 < argc ) cout << endl;
	}	    
    }
    else {
 	if ( wm.empty() )
	    if ( !filemap(default_dict, wm, cm, maxl, minl) )
		cerr << "Can't open " << default_dict<< endl;
    
	for ( i = optind; i < argc ; i++ ) {
	    const wordmaprange none(wm.end(), wm.end()); // a null wp entry
	    wordmatrix wp(wm,pl,0); // The map from pos, len to words for
	                        // this string (because we set everything we
	                        // need, we clear do not clear the matrix
	                        // initially
	    s = argv[i];
	    if ( s.length() > pl ) s = s.substr(0,pl);

	    // The compiler should inline s.length() making it just as
	    // fast as moving it out of these loops.  Confirmed under
	    // egcs-1.1.2.
	
	    // Calculate the pos,length substring mapping for this string
	    for ( start = 0; start < s.length(); start++ )
		for (len = 1;len <= (int) (s.length()-start); len++ ) {
		    w = wm.equal_range(s.substr(start,len));
		    if ( w.first != wm.end() ) 
			wp.set(start, len, w);
		    else
			wp.set(start,len, none);
		}
	    phone2words(s,"",0,wp,words);
	    copy(words.begin(), words.end(),
		 ostream_iterator<const string>(cout, "\n"));
	    words.clear();
	    if ( i+1 < argc ) cout << endl;
	}
    }
    exit(0);
}
