// wordmap and wordmaprange are typedefed in phone.cc

// This implements a mapping between the starting position and length
// of a substring and a range of iterators that point to the vaild
// words that they decode to in the wordmap.  It's a class because
// that makes variable sized 2d arrays easier to code.  A wordmatrix
// is always relative to the wordmap that it provides iterators into,
// so initialization and clearing are always relative to a wordmap.
// clearing a wordmatrix using a different wordmap effectively
// reassigns it to that map.

class wordmatrix {
protected:
    int size;
    wordmaprange **matrix;
public:
    wordmatrix(const wordmap &wm, int sz=maxphonelen, bool cl=1) {
	matrix = new wordmaprange *[size=sz];
	for ( int i = 0; i < size; i++ ) {
	    matrix[i] = new wordmaprange[size+1];
	}
	if ( cl ) clear(wm);
    }
    
    ~wordmatrix() {
	if ( matrix) {
	    for ( int i = 0; i < size; i++ )
		if (matrix[i])
		    delete[] matrix[i];
	    delete[] matrix;
	}
    }
    void clear(const wordmap wm) {
	for ( int i = 0; i < size; i++ ) 
	    for ( int j = 0; j < size+1; j++ )
		matrix[i][j] = wordmaprange(wm.end(),wm.end());
    }

    inline void safeset(const int s, const int l, const wordmaprange &w) {
	if ( s >= 0 && s < size && l >= 0 && l< size+1 ) 
	    matrix[s][l] = w;
	else {
	    cerr << "out of range wordmatrix set (" << s ;
	    cerr << ", " << l << ")" << endl;
	}
    }
    
    inline wordmaprange& safeget(const int s, const int l) const {
	if ( s >= 0 && s < size && l >= 0 && l< size+1 ) 
	    return matrix[s][l];
	else {
	    cerr << "out of range wordmatrix get (" << s ;
	    cerr << ", " << l << ")" << endl;
	    return matrix[0][0];
	}
    }

    inline void set(const int s, const int l, const wordmaprange &w) {
	matrix[s][l] = w;
    }
    
    inline wordmaprange& get(const int s, const int l) const {
	return matrix[s][l];
    }
};

